#Final Project CHE210D
#Student: Bryan R. Goldsmith
#Due  6/14/12

#import python modules
import numpy as np
import time
import atomwrite
import sys
#import compiled fortran library
import finallib

#whether or not to use the 3d realtime visualization
UseVisual = False
if UseVisual:
    #import custom visualization library
    import atomvis

#set the random number seed; useful for debugging
np.random.seed()

#global vars
#Length of box
L = 10.
#Box volume
V = L*L*L
#Number of A particle 
NA = 75
#Numer of B particle 
NB = 75
#Number of total atoms in the system
N = NB + NA
#non bonded interaction cutoff distance is half box length
Cut = L * 0.5

#NOTE:
#everything below assumes unit atomic masses,
#such that forces = accelerations.

def LineSearch(Pos, Walls, Dir, dx, EFracTol,
               Accel = 1.5, MaxInc = 10., MaxIter = 10000):
    """Performs a line search along direction Dir.
Input:
    Pos: starting positions, (N,3) array
    Dir: (N,3) array of gradient direction
    dx: initial step amount
    EFracTol: fractional energy tolerance
    Accel: acceleration factor
    MaxInc: the maximum increase in energy for bracketing
    MaxIter: maximum number of iteration steps
Output:
    PEnergy: value of potential energy at minimum along Dir
    Pos: minimum energy (N,3) position array along Dir
    Walls: Explicit, constrained wall atoms.
"""
    #identify global variables
    global L, Cut, NA
    
    #start the iteration counter
    Iter = 0

    #find the normalized direction    
    NormDir = np.clip(Dir, -1.e100, 1.e100)
    NormDir = NormDir / np.sqrt(np.sum(NormDir * NormDir))

    #take the first two steps and compute energies    
    Dists = [0., dx]
    PEs = [finallib.calcenergy(Pos + NormDir * x, Walls, L, Cut, NA) for x in Dists]
    
    #if the second point is not downhill in energy, back
    #off and take a shorter step until we find one
    while PEs[1] > PEs[0]:
        Iter += 1
        dx = dx * 0.5
        Dists[1] = dx
        PEs[1] = finallib.calcenergy(Pos + NormDir * dx, Walls, L, Cut, NA)
        
    #find a third point
    Dists = Dists + [2. * dx]
    PEs = PEs + [finallib.calcenergy(Pos + NormDir * 2. * dx, Walls, L, Cut, NA)]
    
    #keep stepping forward until the third point is higher
    #in energy; then we have bracketed a minimum
    while PEs[2] < PEs[1]:
        Iter += 1
            
        #find a fourth point and evaluate energy
        Dists = Dists + [Dists[-1] + dx]
        PEs = PEs + [finallib.calcenergy(Pos + NormDir * Dists[-1], Walls, L, Cut, NA)]

        #check if we increased too much in energy; if so, back off
        if (PEs[3] - PEs[0]) > MaxInc * (PEs[0] - PEs[2]):
            PEs = PEs[:3]
            Dists = Dists[:3]
            dx = dx * 0.5
        else:
            #shift all of the points over
            PEs = PEs[-3:]
            Dists = Dists[-3:]
            dx = dx * Accel
            
    #we've bracketed a minimum; now we want to find it to high
    #accuracy
    OldPE3 = 1.e300
    while True:
        Iter += 1
        if Iter > MaxIter:
            print "Warning: maximum number of iterations reached in line search."
            break
            
        #store distances for ease of code-reading
        d0, d1, d2 = Dists
        PE0, PE1, PE2 = PEs

        #use a parobolic approximation to estimate the location
        #of the minimum
        d10 = d0 - d1
        d12 = d2 - d1
        Num = d12*d12*(PE0-PE1) - d10*d10*(PE2-PE1)
        Dem = d12*(PE0-PE1) - d10*(PE2-PE1)
        if Dem == 0:
            #parabolic extrapolation won't work; set new dist = 0 
            d3 = 0
        else:
            #location of parabolic minimum
            d3 = d1 + 0.5 * Num / Dem
            
        #compute the new potential energy
        PE3 = finallib.calcenergy(Pos + NormDir * d3,  Walls, L, Cut, NA)
        
        #sometimes the parabolic approximation can fail;
        #check if d3 is out of range < d0 or > d2 or the new energy is higher
        if d3 < d0 or d3 > d2 or PE3 > PE0 or PE3 > PE1 or PE3 > PE2:
            #instead, just compute the new distance by bisecting two
            #of the existing points along the line search
            if abs(d2 - d1) > abs(d0 - d1):
                d3 = 0.5 * (d2 + d1)
            else:
                d3 = 0.5 * (d0 + d1)
            PE3 = finallib.calcenergy(Pos + NormDir * d3, Walls, L, Cut, NA)
            
        #decide which three points to keep; we want to keep
        #the three that are closest to the minimum
        if d3 < d1:
            if PE3 < PE1:
                #get rid of point 2
                Dists, PEs = [d0, d3, d1], [PE0, PE3, PE1]
            else:
                #get rid of point 0
                Dists, PEs = [d3, d1, d2], [PE3, PE1, PE2]
        else:
            if PE3 < PE1:
                #get rid of point 0
                Dists, PEs = [d1, d3, d2], [PE1, PE3, PE2]
            else:
                #get rid of point 2
                Dists, PEs = [d0, d1, d3], [PE0, PE1, PE3]
                
        #check how much we've changed
        if abs(OldPE3 - PE3) < EFracTol * abs(PE3):
            #the fractional change is less than the tolerance,
            #so we are done and can exit the loop
            break
        OldPE3 = PE3

    #return the position array at the minimum (point 1)        
    PosMin = Pos + NormDir * Dists[1]
    PEMin = PEs[1]

    #if using visualization, update the display
    if UseVisual:
        if atomvis.Initialized:
            #update the positions
            atomvis.Update(PosMin)
        else:
            #initialize the visualization window
            atomvis.Init(PosMin)
        
    return PEMin, PosMin

        
def ConjugateGradient(Pos, Walls, dx, EFracTolLS, EFracTolCG):
    """Performs a conjugate gradient search.
Input:
    Pos: starting positions, (N,3) array
    dx: initial step amount
    EFracTolLS: fractional energy tolerance for line search
    EFracTolCG: fractional energy tolerance for conjugate gradient
Output:
    PEnergy: value of potential energy at minimum
    Pos: minimum energy (N,3) position array 
    Walls: constrained wall atom positions
"""
    #identify global variables
    global L, Cut, NA
    #initial search direction
    Forces = np.zeros_like(Pos)
    PE, Forces = finallib.calcenergyforces(Pos, Walls, L, Cut, NA, Forces)
    Dir = Forces
    OldPE = 1.e300
    #iterative line searches
    while abs(PE - OldPE) > EFracTolCG * abs(PE):
        OldPE = PE
        PE, Pos = LineSearch(Pos, Walls, Dir, dx, EFracTolLS)
        OldForces = Forces.copy()
        PE, Forces = finallib.calcenergyforces(Pos, Walls, L, Cut, NA, Forces)
        Gamma = np.sum((Forces - OldForces) * Forces) / np.sum(OldForces * OldForces)
        Dir = Forces + Gamma *  Dir
    return PE, Pos

def InitWallPositions(L):
    """Returns an array of initial positions of each atom,
placed on a FCC lattice.
Input:
    L: box length
Output:
    Walls: (Nwall,3) array of positions
"""
    #compute integer grid # of locations for cubic lattice
    NLat = 10
    LatSpace = L / NLat
    r = LatSpace * np.arange(NLat, dtype=float) - 0.5*L
	Walls = np.zeros((4.*pow(len(r),2),3),float)
    Wall_z = np.max(r) + LatSpace
    #Empty vector to hold wall names for pdb file
    AtomNames=[]    
    j=0
    #loop through x, y, z positions for all atoms in lattice
    for x in r:
        for y in r:
            #Walls are periodic in x and y direction
            Walls[j]=np.array([x,y,Wall_z+LatSpace],float) #Top and bottom walls with two atomic layers each
            j+=1
            AtomNames.append('C')
            Walls[j]=np.array([x,y,Wall_z],float) 
            j+=1
            AtomNames.append('C')
            Walls[j]=np.array([x,y,-Wall_z],float)
            j+=1
            AtomNames.append('C')
            Walls[j]=np.array([x,y,-Wall_z-LatSpace],float) 
            j+=1
            AtomNames.append('C')
 			
    return Walls, AtomNames
	
def InitPositions(N, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    Names=[]  
    #make the position array
    Pos = np.zeros((N,3), float)
    #compute integer grid # of locations for cubic lattice
    NLat = int(N**(1./3.) + 1.)
    LatSpace = L / NLat
    #make an array of lattice sites
    r = LatSpace/2.5 * np.arange(NLat, dtype=float) - L/6.
    #loop through x, y, z positions in lattice until done
    #for every atom in the system
    i = 0
    for x in r:
        for y in r:
            for z in r:
                Pos[i] = np.array([x,y,z], float)
				#A atoms are stored first
                if (i < NA):
                    Names.append('He')
				#B atom are stored second. Add psuedo-particle names to faciliate CHimera
                else:
                    Names.append('Ar')
                #add a random offset to help initial minimization
                Offset = 0.1 * LatSpace * (np.random.rand(3) - 0.5)
                Pos[i] = Pos[i] + Offset
                i += 1
                #if done placing atoms, return
                if i >= N:
                    return Pos, Names
    return Pos, Names
	
def AndersonThermostat(Vel, T):
    #Anderson thermostat using massive collisions
    sqrT = np.sqrt(T)
    Vel = np.random.normal(0,sqrT,(len(Vel),3)) #could have inverse sampled from uniform distribution
    return Vel

def InitVelocities(N, T):
    """Returns an initial random velocity set.
Input:
    N: number of atoms
    T: target temperature
Output:
    Vel: (N,3) array of atomic velocities
"""
    Vel = np.random.rand(N, 3)
    Vel = AndersonThermostat(Vel, T)
    return Vel

def InitAccel(Pos, Walls):
    """Returns the initial acceleration array.
Input:
    Pos: (N,3) array of atomic positions
Output:
    Accel: (N,3) array of acceleration vectors
"""
    #global variables
    global  L, Cut, NA
    #get the acceleration from the forces
    Forces = np.zeros_like(Pos)
    PEnergy, Accel = finallib.calcenergyforces(Pos, Walls, L, Cut, NA, Forces)
    return Accel

def InstTemp(Vel):
    """Returns the instantaneous temperature.
Input:
    Vel: (N,3) array of atomic velocities
Output:
    Tinst: float
"""
    return np.sum(Vel * Vel) / (3. * len(Vel))

def Histogram(Pos, nbin):
    """Returns a histogram of the z-positions for atoms
Input:
    Pos: (N,3) array of atomic positions
    L: Box length
    nbin: number of bins
Output:
    zB: Histogram of z positions for the B particle
    zA: Histogram of z positions for the A particle
"""
	#global variables
    global  L, Cut, NA
    dz = L/float(nbin) #bin width
    bin = [((-L/2)+i*dz) for i in range(nbin+1)]
    zB = np.zeros((len(bin)-1), int) #make arrays to hold values
    zA = np.zeros((len(bin)-1), int)
    i = 0
	#look at z position of atoms 
    for x in Pos:
        i+=1
        if (i <= NA): #For A atoms
            k=0
            while k<(len(bin)-1): 
                if x[2]>=bin[k] and x[2]<bin[k+1]: #check to see if the z-position is within a particular bin
                    zA[k]+=1 #add a count in bin k
                k+=1 
        else:        #Similiarly for B atoms
            k=0
            while k<(len(bin)-1):
                if x[2]>=bin[k] and x[2]<bin[k+1]:
                      zB[k]+=1
                k+=1
    return bin, zB, zA
	
#Main code

#Define conjugant gradient criteria
EFracTolLS = 1.e-8
EFracTolCG = 1.e-10
dx = 0.001
#Time step
dt = 0.00005
#Step counter
i = 0
j = 0
#Initial md time
t = 0
#Reduced temperature
T = 7.988
#Set frequency of md steps to rescale particle velocities via massive collision
RescaleFreq = 200
#Set frequency to record md steps to pdb file
PdbFreq = 100
#Take a sample point every 1000 md steps during production run
sample = 1000
#Take a sample every 100 md steps during equilibration
sampleeq = 100
#define total energy array and initialize counters to 0
TE = np.zeros((500),float)
#equilibration times 1 and 2, and # of production steps
NStepsEquil = 50000
NStepsEquil2 = 50000
NStepsProd = 500000

#Place atoms on a cubic lattice
(Pos, Patoms) = InitPositions(N, L)

#Generate slit pore walls
(Walls, Watoms) = InitWallPositions(L)

#Create a file for constrained wall positions
PdbC = atomwrite.pdbfile("Walls.pdb", L, AtomNames = Watoms)
PdbC.write(Walls)
PdbC.close()
sys.stdout.flush()

#Histogram declaration
TotalHistcount = 0
bins = 20 #Number of bins in the z-position histogram to calculate # density from wall
(bin, zB, zA) = Histogram(Pos, bins)

if UseVisual:
    atomvis.Init(Pos,L)
sys.stdout.flush()

#print out atoms before conjugate gradient minimization so can check lattice
atomwrite.pdbfile("BeforeCG.pdb", L, AtomNames=Patoms).write(Pos)
	
#minimize energy of starting particle configuration
PEnergy, Pos = ConjugateGradient(Pos, Walls, dx, EFracTolLS, EFracTolCG)

#print out atoms after conjugate gradient minimization so can check lattice
atomwrite.pdbfile("AfterCG.pdb", L, AtomNames=Patoms).write(Pos)

#Calculate initial random velocities scaled to satisfy T*
Vel = InitVelocities(N, T)

#Calculate initial accelerations 
Accel = InitAccel(Pos, Walls)

#Equilibration period 1 w/ anderson thermostat
print "Equilibration period 1"
print "Potential Energy"
while i < NStepsEquil:
    Pos, Vel, Accel, KEnergy, PEnergy =  finallib.vvintegrate(Pos, Walls, Vel, Accel, L, Cut, NA, dt)
    i+=1
    t+=dt
    if i % RescaleFreq == 0:
        Vel = AndersonThermostat(Vel, T)
        print PEnergy
        if UseVisual:
           atomvis.Update(Pos,L)
sys.stdout.flush()		
   
#Equilibration period 2 w/ Anderson thermostat
print "Equilibration period 2"
print "KE, PE, Total energy"
while i < NStepsEquil2:
    Pos, Vel, Accel, KEnergy, PEnergy =  finallib.vvintegrate(Pos, Walls, Vel, Accel, L, Cut, NA, dt)
    i+=1
    t+=dt
    if i % RescaleFreq == 0:
        Vel = AndersonThermostat(Vel, T)
    if i % sampleeq == 0:
        #check energy conservation
        TE[j] = PEnergy + KEnergy
        print " %11.8f, %11.8f, %11.8f" % (KEnergy, PEnergy, TE[j])
        j+=1
        if UseVisual:
           atomvis.Update(Pos,L)
		   		   
#Rescale velocities and begin NVE simulation
TEmean =  TE.mean(axis=0)
KEnergy = TEmean - PEnergy
Vscale = np.sqrt(2*KEnergy/(np.sum(Vel*Vel)))
Vel = Vel*Vscale
#I checked to verify that rescaled velocities give back scaled KEnergy

#copy initial equilibrated positions
Pos0 = Pos.copy()

#Write particle positions to pdb file
Pdb = atomwrite.pdbfile("movie.pdb", L, AtomNames = Patoms)

#reinitialize counters
i=0
t=0
j=0

#define mean squared displacement (MSD), time array, and total energy 
DB = np.zeros((500),float) #B MSD
DA = np.zeros((500),float) #A MSD
t_array = np.zeros((500),float)
TE = np.zeros((500),float) #total energy array
#Empty vectors to fill with histogram values
Bhist = np.zeros_like(zB)  
Ahist = np.zeros_like(zA) 

print "Production run"
print "DA[j], DBj], TE[j])"
#Production run in NVE ensemble
while i < NStepsProd:
    Pos, Vel, Accel, KEnergy, PEnergy =  finallib.vvintegrate(Pos, Walls, Vel, Accel, L, Cut, NA, dt)
    i+=1
    t+=dt
    if i % sample == 0:
	    #check energy conservation
        TE[j] = PEnergy + KEnergy
        #compute MSD
        DA[j] = np.sum((Pos[:NA] -Pos0[:NA] )**2)/len(Pos[:NA])
        DB[j] = np.sum((Pos[NB:] -Pos0[NB:] )**2)/len(Pos[NB:])
        t_array[j] = t
        print "%11.8f,%11.8f,%11.8f" % (DA[j], DB[j], TE[j])
        j+=1
        if i % PdbFreq == 0:
            Pdb.write(Pos)
            if UseVisual:
                atomvis.Update(Pos,L)
		#Count particle distribution in slit pore
        (bin, zB, zA) = Histogram(Pos, bins)
        Bhist = Bhist + zB
        Ahist = Ahist + zA
        TotalHistcount+=1	 #total histogram counter
					
				
#Compute system properties

#compute diffusion coefficients
Slope, Intercept = np.polyfit(t_array, DA , 1)
print "diffusion coefficient A, %11.8f %11.8f" % (1./6.*Slope, T)
Slope, Intercept = np.polyfit(t_array, DB, 1)
print "diffusion coefficient B, %11.8f %11.8f" % (1./6.*Slope, T)

#Compute total energy mean and associated fluctations  for debugging purposes              
TEstd = TE.std(axis=0)
TEmean = TE.mean(axis=0)
print "Tstd/TEmean %11.8f %.4f" % (TEstd/TEmean, dt)
Pdb.close() 

#Make density histograms and print

#Print bin dimensions
print "Bins of Histogram"
w = 0
for x in Ahist:
    print "%1.1f" % ((bin[w]+bin[w+1])/2.)
    w+=1
Bhist = Bhist/(TotalHistcount*V/float(bins)) 
Ahist = Ahist/(TotalHistcount*V/float(bins)) 
print "A histogram"
for x in Ahist:
	print x
print "B histogram"
for x in Bhist:
	print x 	